"""
run system commands easily

see primary run() function
"""
import asyncio
from contextlib import contextmanager
import os
from pathlib import Path
import subprocess
import sys
import shlex
try:
    from shlex import join
except ImportError:
    from subprocess import list2cmdline as join

from packaging import version

from apkg import ex
from apkg.log import getLogger, T, COMMAND, get_log_level


log = getLogger(__name__)


IS_ROOT = os.geteuid() == 0
PY_VERSION = version.parse('%s.%s' % (sys.version_info[:2]))
# Require python 3.7 for tee feature.
CAN_TEE = PY_VERSION >= version.parse('3.7')


def run(cmd,
        *args,
        check=True,
        direct=False,
        tee='auto',
        quiet=False,
        log_fun=log.command,
        **kwargs):
    """
    subprocess.run wrapper with tee and logging powers

    You can use this in following ways:

        run('command', 'arg1', 'arg2')
        run(['command', 'arg1', 'arg2'])
        run('command arg1 arg2', shell=True)

    Differences from subprocess.run:

    * can both capture and output stdout/stderr when tee is set (asyncio)
    * return str subclass with CompletedProcess args - easy to process
    * log commands by default for easy debugging
    * check=True by default - failing commands will raise ex.CommandFailed
    * ex.CommandFailed and ex.CommandNotFound can be raised

    Params:
        cmd: command to run - a str or a List[str]
        *args: optional List[str] of command arguments
        check: raise ex.CommandFailed on command failure
        direct: direct output mode - don't capture stdout and stderr
        tee: both capture and output stdout/stderr (using asyncio)
        quiet: disable tee and command logging
        log_fun: function to log command with (None to disable)

    Return:
        CommandOutput is a str subclass with subprocess.ProcessCompleted
        args allowing for direct string processing with optional access to
        process information:

        out = run('echo', 'hello world')

        assert out == 'hello world'
        print(f'command: {out.args_str})
        print(f'return code: {out.returncode}')
        print(f'stdout:\n{out}')
        print(f'stderr:\n{out.stderr}')
    """
    cmd = parse_cmd_args(cmd, *args)
    shell = kwargs.get('shell', False)
    if shell:
        cmd_str = cmd[0]
    else:
        cmd_str = join(cmd)

    if quiet:
        log_fun = log.verbose_command
        tee = False
    elif tee == 'auto':
        if direct:
            tee = False
        else:
            tee = bool(get_log_level() <= COMMAND)

    if tee and direct:
        log.warning(
            "Running command with both direct=True and tee=True doesn't"
            " make sense. Disabling tee in favor of direct output.")
        tee = False

    if tee and not CAN_TEE:
        log.warning("can't tee on python < 3.7"
                    " - command output won't be printed")
        tee = False

    if log_fun:
        log_fun(cmd_str)

    if tee:
        try:
            result = asyncio.run(_tee(*cmd, **kwargs))
        except FileNotFoundError:
            raise ex.CommandNotFound(cmd=cmd_str)
    else:
        if direct:
            # reuse host process stdout/stderr directly
            stdout, stderr = None, None
        else:
            # this is quivalent of capture_output=True,
            # but we support old pythons
            stdout, stderr = subprocess.PIPE, subprocess.PIPE
        try:
            result = subprocess.run(
                cmd,
                stdout=stdout,
                stderr=stderr,
                check=False,
                universal_newlines=True,
                **kwargs)
        except OSError:
            raise ex.CommandNotFound(cmd=cmd_str)

    cmdout = CommandOutput(result, shell=shell)

    if check and result.returncode != 0:
        raise ex.CommandFailed(cmdout=cmdout)

    return cmdout


async def _tee(*args, shell=False, **kwargs):
    """
    async version of subprocess.run() which can both
    stream and capture stdout/stderr like unix tee

    Print both stdout/stderr to stderr in order not to polute
    stdout with random command output.

    Use run() function from this module with tee=True
    to use this in a convenient way.
    """
    if shell:
        process = await asyncio.create_subprocess_shell(
            args[0],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            **kwargs
        )
    else:
        process = await asyncio.create_subprocess_exec(
            *args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            **kwargs
        )
    out = []
    err = []

    def tee_fun(line, sink):
        line_str = line.decode("utf-8").rstrip()
        sink.append(line_str)
        print(line_str, file=sys.stderr)

    loop = asyncio.get_event_loop()
    tasks = []
    if process.stdout:
        tasks.append(loop.create_task(_read_stream(
            process.stdout, lambda x: tee_fun(x, out))))
    if process.stderr:
        tasks.append(loop.create_task(_read_stream(
            process.stderr, lambda x: tee_fun(x, err))))

    await asyncio.wait(set(tasks))

    stdout = os.linesep.join(out) + os.linesep
    stderr = os.linesep.join(err) + os.linesep

    return subprocess.CompletedProcess(
        args=list(args),
        returncode=await process.wait(),
        stdout=stdout,
        stderr=stderr,
    )


async def _read_stream(stream, callback):
    while True:
        line = await stream.readline()
        if line:
            callback(line)
        else:
            break


def sudo(*cmd, **kwargs):
    """
    run command with sudo
    """
    preserve_env = kwargs.pop('preserve_env', False)
    if 'env' in kwargs:
        preserve_env = True
    if not IS_ROOT:
        shell = kwargs.get('shell', False)
        sudo_cmd = ['sudo']
        if preserve_env:
            sudo_cmd.append('-E')
        if shell:
            shcmd = shlex.quote(cmd[0])
            cmd = ["%s bash -c %s" % (join(sudo_cmd), shcmd)]
        else:
            cmd = parse_cmd_args(*cmd)
            cmd = sudo_cmd + cmd
        if 'log_fun' not in kwargs:
            kwargs['log_fun'] = log.sudo
    return run(*cmd, **kwargs)


def parse_cmd_args(cmd, *args):
    if isinstance(cmd, (str, Path)):
        cmd = [cmd]
    if args:
        cmd.extend(args)
    # convert Path and others to str
    cmd = [str(c) for c in cmd]
    return cmd


@contextmanager
def cd(newdir):
    """
    Temporarily change current directory.
    """
    olddir = os.getcwd()
    oldpwd = os.environ.get('PWD')
    newpath = os.path.abspath(os.path.expanduser(str(newdir)))
    os.chdir(newpath)
    os.environ['PWD'] = newpath
    try:
        yield
    finally:
        os.chdir(olddir)
        if oldpwd is not None:
            os.environ['PWD'] = oldpwd
        else:
            del os.environ['PWD']


class CommandOutput(str):
    """
    A str subclass with CompletedProcess args

    Useful for returning command stdout for easy processing
    while preserving command run information

    Additionally, args_str is available with command args
    properly joined into a string.
    """
    # pylint: disable=unused-argument
    def __new__(cls, result, shell):
        out = (result.stdout or '').rstrip()
        return str.__new__(cls, out)

    def __init__(self, result, shell=False):
        self.args = result.args
        self.shell = shell
        if shell:
            self.args_str = self.args[0]
        else:
            self.args_str = join(self.args)
        self.returncode = result.returncode
        self.stdout = self
        self.stderr = (result.stderr or '').rstrip()
        str.__init__(self)

    @property
    def success(self):
        return self.returncode == 0


class ShellCommand:
    command = None

    def __init__(self):
        if self.command is None:
            self.command = self.__class__.__name__.lower()

    def __call__(self, *params, **kwargs):
        return run(self.command, *params, **kwargs)


def log_cmd_fail(cmdout):
    log.error("command failed: %s", T.command(cmdout.args_str))
    if cmdout.stdout:
        log.bold("stdout:")
        log.info(cmdout.stdout)
    if cmdout.stderr:
        log.bold("stderr:")
        log.info(cmdout.stderr)



#!/usr/bin/env python3
import os
import setuptools


def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string in: %s" % rel_path)


# see pyproject.toml and setup.cfg (legacy) for declarative values for setuptools
setuptools.setup(
    name='apkg',
    version=get_version("apkg/__init__.py"),
)



import click

from apkg import adistro
from apkg import compat
from apkg import ex
from apkg.log import getLogger, T
from apkg.project import Project


log = getLogger(__name__)


@click.command(name='status')
@click.option('-d', '--distro',
              help="override target distro  [default: current]")
@click.help_option('-h', '--help', help='show this help')
def cli_status(*args, **kwargs):
    """
    show status of current project
    """
    status(*args, **kwargs)


def status(distro=None):
    """
    show status of current project
    """
    proj = Project(auto_compat=False)

    msg = "project name:            {t.bold}{name}{t.normal}"
    print(msg.format(name=proj.name, t=T))
    msg = "project base path:       {t.bold}{path}{t.normal}"
    print(msg.format(path=proj.path.base.resolve(), t=T))

    msg = "project VCS:             {t.bold}{vcs}{t.normal}"
    print(msg.format(vcs=proj.vcs or 'none', t=T))

    msg = "project config:          {t.bold}{path}{t.normal}"
    if proj.path.config.exists():
        msg += " ({t.green}exists{t.normal})"
    else:
        msg += " ({t.warn}doesn't exist{t.normal})"
    print(msg.format(path=proj.path.config, t=T))

    msg = "project compat level:    {t.bold}{level}{t.normal}"
    level = proj.compat_level
    compat_ok, compat_state = compat.level_status(level)
    if level is None:
        level = 'N/A'
    if compat_ok:
        msg += " ({t.green}{state}{t.normal})"
    else:
        msg += (" ({t.warn}{state}"
                " -> run {t.command}apkg compat{t.normal})")
    print(msg.format(level=level, state=compat_state, t=T))

    print()
    msg = "package templates path:  {t.bold}{path}{t.normal}"
    if proj.path.templates.exists():
        msg += " ({t.green}exists{t.normal})"
    else:
        msg += " ({t.red}doesn't exist{t.normal})"
    print(msg.format(path=proj.path.templates, t=T))

    print("package templates:")
    if proj.templates:
        msg_lines = []
        for template in proj.templates:
            short_path = template.path.relative_to(proj.path.templates)
            msg_lines.append(
                "    {t.bold}%s{t.normal}: {t.green}%s{t.normal} %s: %s"
                % (short_path, template.pkgstyle.name,
                   template.selection_str(), template.distro_rules))
        msg = "\n".join(msg_lines)
    else:
        msg = "    {t.red}no package templates found{t.normal}"
    print(msg.format(dir=proj.path.templates, t=T))

    print("template variables sources:")
    if proj.variables_sources:
        msg_lines = []
        for vsrc in proj.variables_sources:
            try:
                vsrc.validate()
                line = ("%s: {t.bold}%s{t.normal}"
                        % (vsrc.src_attr, vsrc.src_val))
                if vsrc.exists():
                    line += " ({t.green}exists{t.normal})"
                else:
                    line += " ({t.red}doesn't exist{t.normal})"
                line = line.format(t=T)
            except ex.InvalidVariablesSource as e:
                line = "{t.red}%s{t.normal}".format(t=T) % e
            msg_lines.append("    %s" % line)
        msg = "\n".join(msg_lines)
    else:
        msg = "    no custom variables sources defined"
    print(msg)

    print()
    if distro:
        # target distro status
        distro = adistro.Distro(distro)
        msg = "target distro: {t.cyan}{id}{t.normal}"
        print(msg.format(id=distro, t=T))
    else:
        # current distro status
        distro = adistro.current_distro()
        fullname = adistro.current_fullname()
        msg = ("current distro: "
               "{t.cyan}{id}{t.normal} / {t.cyan}{full}{t.normal}")
        print(msg.format(full=fullname, id=distro, t=T))

    template = proj.get_template_for_distro_(distro)
    msg = "    package style: "
    if template:
        style = template.pkgstyle.name
        msg += "{t.green}%s{t.normal}" % style
    else:
        msg += "{t.warn}unsupported{t.normal}"
    print(msg.format(t=T))
    msg = "    package template: "
    if template:
        msg += "{t.green}%s{t.normal}" % template.path
    else:
        msg += "{t.warn}unsupported{t.normal}"
    print(msg.format(t=T))


APKG_CLI_COMMANDS = [cli_status]



"""
module for handling and rendering apkg package templates
"""
from collections import UserDict
import glob
import importlib.util
import os
import shutil

from pathlib import Path
import jinja2
import jinja2.ext
from markupsafe import Markup

from apkg import ex
from apkg import adistro
from apkg.log import getLogger
from apkg import pkgstyle as _pkgstyle
from apkg.util import common


log = getLogger(__name__)


DUMMY_VARS = {
    'name': 'PKGNAME',
    'version': '0.VERSION',
    'release': '0.RELEASE',
    'nvr': 'NAME-VERSION-RELEASE',
    'distro': adistro.Distro('UNKNOWN')
}

# package template selection types ordered by priority
TS_ALIAS, TS_DISTRO, TS_PKGSTYLE = range(3)
TEMPLATE_SELECTION_STR = {
    TS_ALIAS: 'distro alias',
    TS_DISTRO: 'distro-specific',
    TS_PKGSTYLE: 'pkgstyle default',
}

DEFAULT_IGNORE_FILES = [
    '.*',
]
DEFAULT_PLAIN_COPY_FILES = [
    '*.patch',
]


class IncludeRawExtension(jinja2.ext.Extension):
    """
    custom jinja tag include_raw to insert file contents without templating

    Usage from template:

        {% include_raw 'distro/pkg/foo' %}
    """
    tags = {"include_raw"}

    def parse(self, parser):
        lineno = parser.stream.expect("name:include_raw").lineno
        fn = jinja2.nodes.Const(parser.parse_expression().value)
        result = self.call_method("_render", [fn], lineno=lineno)
        return jinja2.nodes.Output([result], lineno=lineno)

    def _render(self, filename):
        src = self.environment.loader.get_source(self.environment, filename)[0]
        return Markup(src.rstrip('\n'))


# pylint: disable=too-many-instance-attributes
class PackageTemplate:
    def __init__(self, path, style=None, selection=TS_DISTRO,
                 ignore_files=None, plain_copy_files=None,
                 variables_sources=None):
        self.path = Path(path)
        self.style = style
        self.selection = selection
        self.distro_rules = None

        if ignore_files is None:
            self.ignore_files = DEFAULT_IGNORE_FILES
        else:
            self.ignore_files = ignore_files
        if plain_copy_files is None:
            self.plain_copy_files = DEFAULT_PLAIN_COPY_FILES
        else:
            self.plain_copy_files = plain_copy_files
        self.variables_sources = variables_sources

        self.setup_env()

    def setup_env(self):
        self.env = jinja2.Environment(
            loader=jinja2.FileSystemLoader('.'),
            extensions=[IncludeRawExtension])

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, val):
        self._path = Path(val)
        self._name = self._path.name

    @property
    def name(self):
        return self._name

    @property
    def pkgstyle(self):
        if not self.style:
            self.style = _pkgstyle.get_pkgstyle_for_template(self.path)
        return self.style

    def selection_str(self):
        return TEMPLATE_SELECTION_STR.get(self.selection, 'INVALID')

    def match_distro(self, distro):
        return self.distro_rules.match(distro)

    def template_vars(self, tvars=None):
        """
        get/update template variables
        """
        vars_ = DUMMY_VARS.copy()

        # static pkgstyle vars
        vars_pkgstyle = getattr(self.pkgstyle, 'TEMPLATE_VARS', {})
        vars_.update(vars_pkgstyle)
        # dynamic pkgstyle vars resolved at render time
        vars_dyn = getattr(self.pkgstyle, 'TEMPLATE_VARS_DYNAMIC', {})
        if vars_dyn:
            for name, fun in vars_dyn.items():
                vars_[name] = fun()
        # supplied vars
        if tvars:
            vars_.update(tvars)
        # custom variables
        if self.variables_sources:
            vars_ = self.custom_vars(vars_)

        return vars_

    def custom_vars(self, tvars):
        """
        get/update template variables from custom sources
        """
        for vsrc in self.variables_sources:
            vsrc.validate()
            log.info("custom variables from %s: %s",
                     vsrc.src_attr, vsrc.src_val)
            custom_tvars = vsrc.get_variables(tvars)
            log.verbose("%s custom variables: %s",
                        len(custom_tvars), custom_tvars)
            tvars.update(custom_tvars)

        return tvars

    def render(self, out_path, tvars):
        """
        render package template into specified output directory

        Args:
            out_path: output base path
            tvars: variables available from template
        """
        log.info("renderding package template: %s -> %s", self.path, out_path)
        if out_path.exists():
            log.verbose("template render dir exists: %s", out_path)
        else:
            out_path.mkdir(parents=True, exist_ok=True)

        tvars = self.template_vars(tvars=tvars)

        # recursively render all files
        for d, _, files in os.walk(self.path, followlinks=True):
            rel_dir = Path(d).relative_to(self.path)
            dst_dir = out_path / rel_dir
            dst_dir.mkdir(parents=True, exist_ok=True)

            for fn in files:
                dst = out_path / rel_dir / fn
                src = Path(d) / fn
                if common.fnmatch_any(fn, self.ignore_files):
                    log.verbose("ignoring template file: %s", src)
                    continue

                if common.fnmatch_any(fn, self.plain_copy_files):
                    log.verbose(
                        "plain copying file without render: %s -> %s",
                        src, dst)
                    shutil.copyfile(src, dst)
                else:
                    log.verbose("rendering file: %s -> %s", src, dst)
                    t = self.env.get_template(str(src))
                    with dst.open('w') as dstf:
                        dstf.write(t.render(**tvars) + '\n')

                # preserve original permission
                dst.chmod(src.stat().st_mode)

    def render_file_content(self, name, tvars=None):
        """
        render template file in memory and return its content
        """
        src = self.path / name
        tvars = self.template_vars(tvars=tvars)
        t = self.env.get_template(str(src))
        return t.render(**tvars) + '\n'

    def __repr__(self):
        return "PackageTemplate<%s,%s>" % (self.name, self.pkgstyle.name)


def load_templates(path,
                   distro_aliases=None,
                   ignore_files=None,
                   plain_copy_files=None,
                   variables_sources=None):
    """
    load package templates sorted by evaluation priority

    Params:
        path - templates base path (i.e.: distro/pkg)
        distro_aliases - distro aliases dict (optional)
        ignore_files - list of file patterns to ignore (optional)
        plain_copy_files - list of file patterns to copy
                           without templating (optional)
        variables_sources - list of sources to define
                            template variables (optional)

    Returns:
        list of PackageTemplates in order they should
        be evaluated (by selection type):

        1) distro alias templates
        2) distro-specific templates
        3) pkgstyle default templates
    """
    alias_tps = []
    distro_tps = []
    pkgstyle_tps = []

    aliases = distro_aliases or {}

    for entry_path in glob.glob('%s/*' % path):
        if not os.path.isdir(entry_path):
            # ignore non-dirs
            continue

        template = PackageTemplate(
            entry_path,
            ignore_files=ignore_files,
            plain_copy_files=plain_copy_files,
            variables_sources=variables_sources)

        if not template.pkgstyle:
            log.warning("ignoring unknown package style in %s", entry_path)
            continue

        alias_rules = aliases.get(template.name)
        if alias_rules:
            # distro alias template (name match)
            template.selection = TS_ALIAS
            template.distro_rules = alias_rules
            alias_tps.append(template)
            continue

        dstyle = _pkgstyle.PKGSTYLES.get(template.name)
        if dstyle:
            # pkgstyle default template (name match)
            template.selection = TS_PKGSTYLE
            rules = adistro.distro_rules(dstyle.SUPPORTED_DISTROS)
            template.distro_rules = rules
            pkgstyle_tps.append(template)
            continue

        # distro-specific template
        template.distro_rules = adistro.name2rule(template.name)
        distro_tps.append(template)

    alias_tps = sort_alias_templates(alias_tps, distro_aliases)
    distro_tps = adistro.sort_by_name(distro_tps)
    pkgstyle_tps.sort(key=lambda x: x.name)
    templates = alias_tps + distro_tps + pkgstyle_tps
    return templates


def sort_alias_templates(alias_templates, distro_aliases):
    result = []
    for da in distro_aliases:
        for at in alias_templates:
            if at.name == da:
                result.append(at)
                break
    assert len(alias_templates) == len(result)
    return result


class VariablesSource(UserDict):
    """
    template variables from external variables source
    """
    VARS_SOURCES = ['local_module', 'python_module']

    def __init__(self, **kwargs):
        super().__init__(kwargs)

    @property
    def local_module(self):
        return self.get('local_module', None)

    @property
    def python_module(self):
        return self.get('python_module', None)

    @property
    def src_attr(self):
        for src in self.VARS_SOURCES:
            if src in self.data:
                return src
        return None

    @property
    def src_val(self):
        return self.data.get(self.src_attr, None)

    def validate(self):
        src_attrs = [src for src in self.VARS_SOURCES if self.get(src)]
        n_src = len(src_attrs)
        if n_src < 1:
            raise ex.InvalidVariablesSource(
                src="no source specified: %s" % self)
        if n_src > 1:
            raise ex.InvalidVariablesSource(
                src="multiple conflicting sources: %s" % self)
        return True

    def exists(self):
        if self.src_attr == 'local_module':
            return Path(self.local_module).exists()
        elif self.src_attr == 'python_module':
            try:
                spec = importlib.util.find_spec(self.python_module)
                return spec is not None
            except ModuleNotFoundError:
                return False
        else:
            return False

    def get_variables(self, tvars):
        if self.local_module:
            if not os.path.exists(self.local_module):
                raise ex.InvalidVariablesSource(
                    src='local_module not found: %s' % (self.local_module))

            spec = importlib.util.spec_from_file_location(
                "custom_variables", self.local_module)

            custom_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(custom_module)
        elif self.python_module:
            try:
                custom_module = importlib.import_module(self.python_module)
            except ModuleNotFoundError:
                raise ex.InvalidVariablesSource(
                    src='python_module not found: %s' % self.python_module)

        custom_fun = getattr(custom_module, 'get_variables', None)
        if not custom_fun:
            raise ex.InvalidVariablesSource(
                src='%s missing required get_variables function: %s'
                % (self.src_val, custom_module))

        custom_tvars = custom_fun(tvars)
        return custom_tvars


def parse_variables_sources(config):
    sources = []
    for vcfg in config:
        vsrc = VariablesSource(**vcfg)
        sources.append(vsrc)
    return sources



# -*- encoding: utf-8 -*-
import contextlib
import os

from apkg import ex
from apkg.util.run import run
from apkg.util.run import ShellCommand


@contextlib.contextmanager
def git_branch(branch):
    if branch:
        old_branch = git.current_branch()
        git('checkout', branch)
        try:
            yield
        finally:
            git('checkout', old_branch)
    else:
        yield


@contextlib.contextmanager
def git_revision(gitrev):
    if gitrev:
        old_gitrev = git.current_commit()
        git.checkout(gitrev)
        try:
            yield
        finally:
            git.checkout(old_gitrev)
    else:
        yield


class Git(ShellCommand):
    command = "git"

    def __call__(self, *params, **kwargs):
        # allows us to run git in isolated mode, avoiding interaction with user
        # specific config, like ~/.git-templates/hooks (used for testing)
        if kwargs.get('isolated', False):
            env = kwargs.get('env', os.environ.copy())
            env['GIT_CONFIG_NOSYSTEM'] = '1'
            env['GIT_CONFIG_NOGLOBAL'] = '1'
            kwargs['env'] = env
        return run(self.command, *params, **kwargs)

    def create_branch_from_remote(self, branch, remote_branch=None):
        lbr = self.local_branches()
        if branch in lbr or remote_branch in lbr:
            return
        if remote_branch is None:
            for rbr in self.remote_branches():
                br = rbr[rbr.find('/') + 1:]
                if branch in [rbr, br]:
                    remote_branch = rbr
        elif remote_branch not in self.remote_branches():
            raise ex.InvalidUsage(
                msg="Branch %s doesn't exist in remotes" % remote_branch)
        self.create_branch(branch, remote_branch)

    def _parse_output(self, out):
        output = [o.strip() for o in out.split("\n") if o]
        return output

    def _parse_branch_output(self, out):
        output = [o for o in self._parse_output(out) if o.find('HEAD') < 0]
        return output

    def remote_branches(self, remote=""):
        res = self("branch", "-r", "--no-color", quiet=True)
        branches = self._parse_branch_output(res)
        branches = [b.replace("remotes/", "")
                    for b in branches if b.startswith(remote)]
        return branches

    def local_branches(self):
        res = self("branch")
        res = self._parse_branch_output(res)
        res = [b.replace("* ", "") for b in res]
        return res

    def current_branch(self):
        branch = self('rev-parse', '--abbrev-ref', 'HEAD', quiet=True)
        return branch

    def current_commit(self):
        commit = self('rev-parse', 'HEAD', quiet=True)
        return commit

    def current_commit_message(self):
        commit_msg = self('log', '-n1', '--pretty=%B', quiet=True)
        return commit_msg

    def ref_exists(self, ref):
        o = self('show-ref', '--verify', '--quiet', ref,
                 check=False, quiet=True)
        return o.success

    def branch_exists(self, branch):
        return self.ref_exists('refs/heads/%s' % branch)

    def object_type(self, ref):
        o = self('cat-file', '-t', ref,
                 check=False, quiet=True)
        if not o:
            return None
        return o

    def is_clean(self):
        o = self('status', '-uno', '--porcelain', quiet=True)
        if o:
            return False
        return True

    def remotes(self):
        res = self("remote", "show", quiet=True)
        return self._parse_branch_output(res)

    def remote_branch_split(self, branch):
        parts = branch.split('/')
        n_parts = len(parts)
        if n_parts >= 2:
            remotes = self.remotes()
            for n in range(1, n_parts):
                remote = '/'.join(parts[:n])
                if remote in remotes:
                    _branch = '/'.join(parts[n:])
                    return remote, _branch
        return None, branch

    def remote_of_local_branch(self, branch):
        try:
            o = self('for-each-ref', '--format=%(upstream:short)',
                     'refs/heads/%s' % branch, quiet=True)
        except ex.CommandFailed:
            return None
        if not o:
            return None
        return o

    def delete_branch(self, branch):
        if self.branch_exists(branch):
            self('branch', '-D', branch)

    def create_branch(self, new_branch, branch):
        try:
            self('branch', '-f', new_branch, branch)
        except ex.CommandFailed:
            # this could only fail if we're on the branch
            self('reset', '--hard', branch)

    def squash_last(self, branch=None):
        if branch is not None:
            self('checkout', branch)
        self("reset", "--soft", "HEAD~")
        # on git < 1.7.9 --no-edit can be done by:
        # git commit -C HEAD --amend
        self("commit", "--amend", "--no-edit")

    def branch_needs_push(self, branch=None):
        if not branch:
            branch = self.current_branch()
        remote_branch = self.remote_of_local_branch(branch)
        h1 = self.get_latest_commit_hash(branch)
        h2 = self.get_latest_commit_hash(remote_branch)
        return h1 != h2

    def linearize(self, starting_point, branch=None):
        if branch is not None and self.current_branch() != branch:
            self('checkout', branch)
        self('rebase', starting_point)

    def rev_range(self, from_revision, to_revision=None):
        rng = from_revision + '..'
        if to_revision:
            rng += to_revision
        return rng

    def get_commits(self, from_revision, to_revision=None):
        rng = self.rev_range(from_revision, to_revision)
        log_out = self('log', '--format=%h %s', rng, quiet=True)
        commits = self._parse_output(log_out)
        return map(lambda x: tuple(x.split(' ', 1)), commits)

    def get_commit_subjects(self, from_revision, to_revision=None):
        rng = self.rev_range(from_revision, to_revision)
        log_out = self('log', '--format=%s', rng, quiet=True)
        return self._parse_output(log_out)

    def get_commit_hashes(self, from_revision, to_revision=None):
        rng = self.rev_range(from_revision, to_revision)
        log_out = self('log', '--format=%h', rng, quiet=True)
        return self._parse_output(log_out)

    def get_latest_commit_hash(self, ref=None):
        cmd = ['log', '-n', '1', '--format=%H']
        if ref:
            cmd.append(ref)
        out = self(cmd, quiet=True, check=False)
        return out

    def get_commits_of_branch(self, branch=None,
                              master='remotes/upstream/master'):
        cmd = ['log', master + '..' + branch, '--oneline', '--reverse',
               '--format="%H %ct"']
        if not branch:
            return []
        out = self(cmd, quiet=True, check=False)
        out = out.replace('"', '')
        if not out:
            return []
        else:
            commits = []
            for c in out.split('\n'):
                commits.append({'hash': c.split(" ")[0],
                                'timestamp': c.split(" ")[1]})
            return commits

    def get_latest_tag(self, branch=None):
        cmd = ['describe', '--abbrev=0', '--tags']
        if branch:
            cmd.append(branch)
        out = self(cmd, quiet=True)
        return out

    def get_file_authors(self, path, with_email=True):
        if with_email:
            pf = '%an <%ae>'
        else:
            pf = '%an'
        authors = self('log', '--oneline', "--pretty=%s" % pf, path,
                       quiet=True)
        authors = reversed(authors.split('\n'))
        return authors

    def get_file_content(self, rev, path):
        obj = '%s:%s' % (rev, path)
        return self('show', obj, quiet=True)

    def config_get(self, param, default=None):
        '''Return the value of a git configuration option.  This will
        return the value of the default parameter (which defaults to
        None) if the given option does not exist.'''

        try:
            return self("config", "--get", param, quiet=True)
        except ex.CommandFailed:
            return default

    def config_set(self, param, value, is_global=False):
        params = [param, value]
        if is_global:
            params.insert(0, '--global')
        return self("config", *params)

    def checkout(self, branch):
        self("checkout", branch, quiet=True)

    def remove(self, ref):
        self('rebase', '--onto', ref + '^', ref, '--preserve-merges',
             '--committer-date-is-author-date')

    def get_timestamp_by_ref(self, ref):
        timestamp = self('show', '-s', '--oneline', '--format="%ct"',
                         ref, quiet=True)
        return timestamp.replace('"', '')


git = Git()



from contextlib import contextmanager
import fnmatch
import hashlib
from pathlib import Path
import re
import shutil
import sys
import tempfile
from typing import Iterable, Mapping, Optional, Union

import yaml

from apkg import ex
from apkg.log import getLogger


log = getLogger(__name__)


PATH_FORMAT = 'relative'


CacheableEntry = Union[str, int, bool, Path,
                       Iterable['CacheableEntry'],
                       Mapping[str, 'CacheableEntry']]


def copy_paths(cache_entry: CacheableEntry, dst: Path) -> CacheableEntry:
    """
    utility to copy a list of paths to dst
    """
    if not dst.exists():
        dst.mkdir(parents=True, exist_ok=True)
    dst_full = dst.resolve()

    if isinstance(cache_entry, Path):
        if cache_entry.parent.resolve() != dst_full:
            p_dst = dst / cache_entry.name
            log.verbose("copying file: %s -> %s", cache_entry, p_dst)
            shutil.copy(cache_entry, p_dst)
            return p_dst
        return cache_entry
    elif isinstance(cache_entry, list):
        return [copy_paths(p) for p in cache_entry]
    elif isinstance(cache_entry, dict):
        result = {}
        for k, v in cache_entry.items():
            result[k] = copy_paths(v)
        return result
    return cache_entry


def get_cached_paths(proj, cache_key: str,
                     result_dir: Optional[str] = None) -> CacheableEntry:
    """
    get cached files and move them to result_dir if specified
    """
    paths = proj.cache.get(cache_key)
    if not paths:
        return None
    if result_dir:
        paths = copy_paths(paths, Path(result_dir))
    return paths


def print_results(results):
    """
    print results received from apkg command
    """
    try:
        for r in results:
            print(format_path(r))
    except TypeError:
        print(str(results))


def print_results_dict(results):
    """
    print results dict as YAML (used in make-archive and get-archive)
    """
    print(yaml_dump(results))


def format_path(path: Path):
    if PATH_FORMAT == 'absolute':
        return str(path.absolute())
    elif PATH_FORMAT == 'stem':
        return str(path.stem)
    return str(path)


def yaml_path_representer(dumper, obj):
    # to print pathlib.Path as str
    return dumper.represent_scalar("tag:yaml.org,2002:str", format_path(obj))


class SafeDumper(yaml.dumper.SafeDumper):
    # don't modify global PyYAML state
    pass


yaml.add_representer(
    # print pathlib.Path as str
    type(Path()),
    yaml_path_representer,
    SafeDumper,
)


def yaml_dump(*args, **kwargs):
    kwargs['Dumper'] = SafeDumper
    return yaml.dump(*args, **kwargs).rstrip()


def parse_inputs(inputs, in_files, in_format='list'):
    """
    utility parser of apkg command inputs
    """
    all_inputs = list(inputs) if inputs else []

    if in_files:
        if len([fl for fl in in_files if fl == '-']) > 1:
            fail = "requested to read stdin multiple times"
            raise ex.InvalidInput(fail=fail)

        for fl in in_files:
            if fl == '-':
                f = sys.stdin
            else:
                f = open(fl, 'r', encoding='utf-8')
            all_inputs += [ln.rstrip() for ln in f.readlines()]
            f.close()

    if in_format == 'yaml':
        result = parse_yaml_inputs(all_inputs)
    else:
        result = parse_list_inputs(all_inputs)

    return result


def parse_list_inputs(inputs):
    return [Path(i) for i in inputs]


def parse_yaml_inputs(inputs):
    if not inputs:
        return {}
    txt = '\n'.join(inputs)
    result = yaml.safe_load(txt)
    return result


def ensure_inputs(inputs, n=0):
    if not inputs:
        raise ex.InvalidInput(
            fail="no input file specified")
    if n:
        n_in = len(inputs)
        if n_in != n:
            exp = 'single input file' if n == 1 else '%s input files' % n
            ins = '\n'.join([str(p) for p in inputs])
            raise ex.InvalidInput(
                fail="expected %s, but got %s:\n\n%s" % (exp, n_in, ins))
    for f in inputs:
        if not f or not f.exists():
            raise ex.InvalidInput(
                fail="input file not found: %s" % f)


@contextmanager
def text_tempfile(text, prefix='apkg_tmp_'):
    """
    write text to a new temporary file and return its path

    file is deleted after use
    """
    f = tempfile.NamedTemporaryFile(
        prefix=prefix, mode='w+t', delete=False)
    path = Path(f.name)
    f.write(text)
    f.close()
    try:
        yield path
    finally:
        path.unlink()


def hash_file(*paths, algo='sha256'):
    """
    return hashlib's hash computed over the contents of the specified file

    typical use case: `hash_file('/path').hexdigest()`
    """
    # code based on https://stackoverflow.com/a/44873382/587396
    h = getattr(hashlib, algo)()
    b = bytearray(128*1024)
    mv = memoryview(b)
    for path in paths:
        with open(path, 'rb', buffering=0) as f:
            while True:
                # NOTE: pylint's cell-var-from-loop cries made me do this >:(
                n = f.readinto(mv)
                if n == 0:
                    break
                h.update(mv[:n])
    return h


def hash_path(*paths, algo='sha256'):
    """
    return hashlib's hash computed over the supplied file paths (as strings)

    typical use case: `hash_path('/path').hexdigest()`
    """
    h = getattr(hashlib, algo)()
    for path in paths:
        h.update(str(path).encode('utf-8'))
    return h


def fnmatch_any(filename, patterns):
    """
    check if a filename matches any of supplied patterns
    """
    for p in patterns:
        if fnmatch.fnmatch(filename, p):
            return True
    return False


def serialize(obj):
    if isinstance(obj, (list, tuple)):
        return [serialize(v) for v in obj]
    if isinstance(obj, dict):
        return {k: serialize(v) for k, v in obj.items()}
    if isinstance(obj, (str, bool, int, float)):
        return obj
    return str(obj)


class SortReversor:
    """
    use this with multi-key sort() to reverse individual keys
    """
    def __init__(self, obj):
        self.obj = obj

    def __eq__(self, other):
        return other.obj == self.obj

    def __lt__(self, other):
        return other.obj < self.obj


def sanitize_fn(name):
    """
    sanitize a string to be safe for use as a filename
    """
    return re.sub(r'[\\/\\:*?"\'<>| ]', '_', name)


def set_path_format(fmt: str) -> str:
    """
    set global path format
    """
    global PATH_FORMAT

    if not fmt or 'relative'.startswith(fmt):
        PATH_FORMAT = 'relative'
    elif 'absolute'.startswith(fmt):
        PATH_FORMAT = 'absolute'
    elif 'stem'.startswith(fmt):
        PATH_FORMAT = 'stem'
    else:
        raise ex.InvalidFormat(
            msg="Invalid path format: %s" % fmt)


def get_path_format():
    """
    set global path format
    """
    return PATH_FORMAT



from pathlib import Path
import shutil

import click

from apkg import adistro
from apkg.cache import file_checksum, path_checksum
from apkg import ex
from apkg.commands.build_dep import build_dep as cmd_build_dep
from apkg.commands.srcpkg import srcpkg as make_srcpkg
from apkg.log import getLogger
from apkg.project import Project
from apkg.util import common


log = getLogger(__name__)


@click.command(name="build")
@click.argument('inputs', nargs=-1)
@click.option('-s', '--srcpkg', is_flag=True,
              help="use source package")
@click.option('-a', '--archive', is_flag=True,
              help="use template (/build srcpkg) from archive")
@click.option('-u', '--upstream', is_flag=True,
              help="use upstream template / archive / srcpkg")
@click.option('-v', '--version',
              help=("upstream archive version to use"
                    ", implies --upstream"
                    ", exclusive with --srcpkg and --archive"))
@click.option('-r', '--release',
              help="set packagge release  [default: 1]")
@click.option('-d', '--distro',
              help="override target distro  [default: current]")
@click.option('-b', '--build-dep', is_flag=True,
              help="install build dependencies on host (apkg build-dep)")
@click.option('-O', '--result-dir',
              help=("put results into specified dir"
                    "  [default: pkg/srcpkg/DISTRO/NVR]"))
@click.option('--cache/--no-cache', default=True, show_default=True,
              help="enable/disable cache")
@click.option('-F', '--in-file', 'in_files', multiple=True,
              help="specify input file, '-' to read from stdin")
@click.option('-I', '--isolated', is_flag=True,
              help="use isolated builder (pbuilder, mock, ...)")
@click.option('-i', '--install-dep', 'build_dep', is_flag=True,
              help="[DEPRECATED] compat alias for --build-dep")
@click.help_option('-h', '--help',
                   help="show this help message")
def cli_build(*args, **kwargs):
    """
    build packages
    """
    results = build(*args, **kwargs)
    common.print_results(results)
    return results


def build(
        srcpkg=False,
        archive=False,
        upstream=False,
        inputs=None,
        in_files=None,
        version=None,
        release=None,
        distro=None,
        result_dir=None,
        build_dep=False,
        isolated=False,
        cache=True,
        project=None):
    """
    build packages
    """
    log.bold('building packages')

    proj = project or Project()
    distro = adistro.distro_arg(distro, proj)
    log.info("target distro: %s", distro)
    use_cache = proj.cache.enabled(
        'local', cmd='build', use_cache=cache)

    inputs = common.parse_inputs(inputs, in_files)

    if build_dep:
        if isolated:
            # doesn't make sense in isolated build
            log.warning("ignoring build-dep request in isolated build")
        else:
            # install build deps if requested
            try:
                cmd_build_dep(
                    srcpkg=srcpkg,
                    archive=archive,
                    upstream=upstream,
                    inputs=inputs,
                    distro=distro,
                    project=proj)
            except ex.DistroNotSupported as e:
                log.warning("SKIPPING build-dep due to error: %s", e)

    if srcpkg:
        if version:
            raise ex.InvalidInput(
                fail="--srcpkg and --version options are mutually exclusive")
    else:
        # make source package
        inputs = make_srcpkg(
            archive=archive,
            inputs=inputs,
            upstream=upstream,
            version=version,
            release=release,
            distro=distro,
            project=proj,
            cache=cache)

    common.ensure_inputs(inputs)
    srcpkg_path = inputs[0]
    if srcpkg:
        log.info("using existing source package: %s", srcpkg_path)

    # fetch pkgstyle (deb, rpm, arch, ...)
    template = proj.get_template_for_distro(distro)
    pkgstyle = template.pkgstyle

    # get needed paths
    nvr = pkgstyle.get_srcpkg_nvr(srcpkg_path)
    build_path = proj.path.package_build / distro.idver / nvr
    if result_dir:
        result_path = Path(result_dir)
    else:
        result_path = proj.path.package_out / distro.idver / nvr

    # check cache
    if use_cache:
        cache_key = 'pkg/%s/%s/' % (distro.idver, nvr)
        cache_key += '%s:%s' % (
            path_checksum(*inputs), file_checksum(*inputs))
        cached = common.get_cached_paths(proj, cache_key, result_dir)
        if cached:
            log.success("reuse %d cached packages", len(cached))
            return cached

    log.info("source package NVR: %s", nvr)
    log.info("build dir: %s", build_path)
    log.info("result dir: %s", result_path)
    # ensure build build doesn't exist
    if build_path.exists():
        log.info("removing existing build dir: %s", build_path)
        shutil.rmtree(build_path)
    # ensure result dir doesn't exist unless specified
    if not result_dir and result_path.exists():
        log.info("removing existing result dir: %s", result_path)
        shutil.rmtree(result_path)

    # build package using chosen distro packaging style
    pkgs = pkgstyle.build_packages(
        build_path,
        result_path,
        srcpkg_paths=inputs,
        isolated=isolated)

    if not pkgs:
        msg = ("package build reported success but there are "
               "no packages:\n\n%s" % result_path)
        raise ex.UnexpectedCommandOutput(msg=msg)
    log.success("built %s packages in: %s", len(pkgs), result_path)

    # sort output for determinism
    pkgs.sort()

    if use_cache and not upstream:
        unfiles = [p for p in pkgs if not p.is_file()]
        if not unfiles:
            # only cache regular files for now
            proj.cache.update(cache_key, pkgs)

    return pkgs


APKG_CLI_COMMANDS = [cli_build]



import json

import click
import distro as distro_mod
import importlib
from pathlib import Path
import sys

from apkg import adistro
from apkg.pkgstyle import PKGSTYLES
from apkg import pkgtemplate
from apkg.log import getLogger, T
from apkg.project import Project
from apkg import terminal
from apkg.util import common
from apkg.util import toml


log = getLogger(__name__)


RUNTIME_DEPS = [
    'bs4',
    'click',
    'distro',
    'jinja2',
    'packaging',
    'requests',
]

BUILD_DEPS = [
    'build',
    'setuptools',
]


@click.group(name='info')
@click.help_option('-h', '--help', help='show command help')
def cli_info():
    """
    show various apkg information
    """


@cli_info.command()
@click.help_option('-h', '--help', help='show command help')
def apkg_deps():
    """
    show apkg dependencies info
    """
    print("{t.bold}Python interpreter{t.normal}:".format(t=T))
    py_cmd = Path(sys.executable).stem
    s = ("{t.green}{c}{t.normal} {t.cyan}{v}{t.normal}: "
         "{t.bold}{p}{t.normal}")
    if sys.prefix != sys.base_prefix:
        s += ' - {t.magenta}venv{t.normal}'
    print(s.format(c=py_cmd, v=sys.version, p=sys.executable, t=T))

    print("\n{t.bold}core dependencies{t.normal}:".format(t=T))
    for dep in RUNTIME_DEPS:
        print(modinfo_t(dep))

    print("\n{t.bold}dynamic dependencies{t.normal}:".format(t=T))
    print(modinfo_t(toml.LOAD_LIB, 'TOML load'))
    print(modinfo_t(toml.DUMP_LIB, 'TOML dump', failc='yellow'))
    print(modinfo_t(terminal.COLOR_LIB, 'terminal colors', failc='yellow'))

    print("\n{t.bold}build dependencies{t.normal}:".format(t=T))
    for dep in BUILD_DEPS:
        print(modinfo_t(dep))


@cli_info.command()
@click.help_option('-h', '--help', help='show command help')
def cache():
    """
    show apkg cache contents
    """
    proj = Project()
    cache_str = "{t.bold}{fn}{t.normal}".format(fn=proj.path.cache, t=T)
    if proj.path.cache.exists():
        log.info("apkg cache: %s", cache_str)
        cdata = json.load(proj.path.cache.open('rt'))
        print(json.dumps(cdata, indent=4))
    else:
        log.info("apkg cache doesn't exist: %s", cache_str)


@cli_info.command()
@click.help_option('-h', '--help', help='show command help')
def config():
    """
    show apkg project configuration
    """
    proj = Project()
    config_str = "{t.bold}{fn}{t.normal}".format(fn=proj.path.config, t=T)
    if proj.path.config.exists():
        log.info("project config: %s\n", config_str)
    else:
        log.info("project config doesn't exist: %s", config_str)

    if proj.config:
        print(toml.dumps(proj.config))


# pylint: disable=redefined-outer-name
@cli_info.command()
@click.help_option('-h', '--help', help='show command help')
def distro():
    """
    show current distro information
    """
    info = distro_mod.info()
    print(toml.dumps(info))


@cli_info.command()
@click.help_option('-h', '--help', help='show command help')
def distro_aliases():
    """
    list available distro aliases
    """
    proj = Project()
    if not proj.distro_aliases:
        log.info("no distro aliases defined")
        return

    for name, al in proj.distro_aliases.items():
        msg = "{t.bold}{name}{t.normal}: {rules}"
        print(msg.format(name=name, rules=al, t=T))


@cli_info.command()
@click.help_option('-h', '--help', help='show command help')
def pkgstyles():
    """
    list available packaging styles
    """
    for name, mod in PKGSTYLES.items():
        print("{t.bold}{name}{t.normal}".format(name=name, t=T))
        msg = "    module:   {t.magenta}{module}{t.normal}"
        print(msg.format(module=mod.__name__, t=T))
        msg = "    file:     {t.magenta}{fn}{t.normal}"
        print(msg.format(fn=mod.__file__, t=T))

        msg = "    distros:  "
        ds = ['{t.bold}%s{t.normal}' % d for d in mod.SUPPORTED_DISTROS]
        msg += ' | '.join(ds)
        print(msg.format(t=T))


@cli_info.command()
@click.option('-d', '--distro',
              help="set target distro  [default: current]")
@click.option('-c', '--custom', is_flag=True,
              help="only show custom variables per source")
@click.help_option('-h', '--help', help='show command help')
def template_variables(distro=None, custom=False):
    """
    show variables available in packaging template
    """
    proj = Project()
    distro = adistro.distro_arg(distro, proj)
    log.info("target distro: %s", distro)
    template = proj.get_template_for_distro(distro)
    if not custom:
        tvars = {'distro': distro}
        tvars = template.template_vars(tvars)
        print(toml.dumps(common.serialize(tvars)))
        return

    # custom variables
    tvars = pkgtemplate.DUMMY_VARS
    tvars['distro'] = distro
    for vsrc in proj.variables_sources:
        print("# variables from %s: %s" % (vsrc.src_attr, vsrc.src_val))
        custom_tvars = vsrc.get_variables(tvars)
        print(toml.dumps(common.serialize(custom_tvars)))
        tvars.update(custom_tvars)


@cli_info.command()
@click.help_option('-h', '--help', help='show command help')
def upstream_version():
    """
    show detected project upstream version
    """
    proj = Project()
    msg = "upstream version: {t.bold}{v}{t.normal}"
    print(msg.format(v=proj.upstream_version, t=T))


def modinfo_t(modname, desc=None, failc='red'):
    if not modname:
        s = "{d}: {t.%s}no supported module found{t.normal}" % (failc)
        s = s.format(d=desc, t=T)
        return s

    mod = None
    err = None
    try:
        mod = importlib.import_module(modname)
    except Exception as e:
        err = e

    if mod:
        s = ''
        if desc:
            s += '%s: ' % desc
        s += "{t.green}{m}{t.normal}"
        if hasattr(mod, '__version__'):
            ver = getattr(mod, '__version__')
            s += " {t.cyan}%s{t.normal}" % ver
        s += ": {t.bold}{p}{t.normal}"
        return s.format(m=modname, p=Path(mod.__file__).parent, t=T)

    s = ''
    if desc:
        s += '%s: ' % desc
    s += ("{t.red}{m}{t.normal}: {t.bold_red}{et}{t.normal}: "
          "{t.normal}{e}{t.normal}")
    return s.format(m=modname, et=err.__class__.__name__, e=err, t=T)


APKG_CLI_COMMANDS = [cli_info]



"""
apkg package style for **Nix** ([NixOS.org](https://nixos.org)).

**source template:**

- `default.nix` as-if in [https://github.com/NixOS/nixpkgs](https://github.com/NixOS/nixpkgs)
- `top-level.nix` that simply wraps it to work outside that official tree,
   in particular it should substitute the source archive;
   e.g. see `apkg/distro/pkg/nix/top-level.nix`

**source package:** the same, just with templates substituted

**packages:** symlink to your local nix store (for the primary package output)
"""
import os
import re
import shutil

from apkg import ex
from apkg.log import getLogger
from apkg.util.run import run
from apkg.util.common import hash_file

log = getLogger(__name__)


SUPPORTED_DISTROS = [
    'nix',
    'nixos',
]


def fname_(path):
    return path / "default.nix"


def is_valid_template(path, distro=None):
    return ((path / "default.nix").exists()
            and (path / "top-level.nix").exists())


def get_template_name(template, distro=None):
    # I'd like to simply use nix directly, e.g.:
    #   return run("nix", "eval", "--file", path / "top-level.nix",
    #           "pname", "--raw")
    # but that would require substituting the templates first.
    # So we use a hacky regexp instead :-/
    expr = fname_(template.path)
    for line in expr.open():
        m = re.match(r'\s*pname\s*=\s*"(\S+)";', line)
        if m:
            return m.group(1)
    raise ex.ParsingFailed(
        msg="unable to determine Name from: %s" % expr)


def get_srcpkg_nvr(path):
    # use source package parent dir as NVR
    return path.resolve().parent.name


def get_build_deps_from_template(  # pylint: disable=unused-argument
        template_path,
        **kwargs):
    # With nix it doesn't make so much sense to care about build deps.
    return []


def install_build_deps(  # pylint: disable=unused-argument
        deps,
        **kwargs):
    return


def build_srcpkg(
        build_path,
        out_path,
        archive_info,
        template,
        tvars):
    archive_path = archive_info['archive']
    tvars = tvars or {}
    tvars['src_hash'] = hash_file(archive_path).hexdigest()
    log.info("applying templates")
    template.render(build_path, tvars=tvars)

    log.info("copying everything to: %s", out_path)
    shutil.copytree(build_path, out_path)
    archive_paths = [archive_path]
    archive_paths.extend(archive_info.get("components", {}).values())
    out_archives = []
    for src_path in archive_paths:
        dst_path = out_path / src_path.name
        shutil.copyfile(src_path, dst_path)
        out_archives.append(dst_path)

    return [out_path / 'top-level.nix', out_path / 'default.nix',
            *out_archives]
    # TODO: maybe list everything in the directory?
    #       (e.g. local patches might be there)


def build_packages(
        build_path,  # pylint: disable=unused-argument
        out_path,
        srcpkg_paths,
        **_):
    log.info("building using nix (silent unless fail)")
    # TODO: use "nix" command CLI after versions < 2.4 get less common
    # (because `--extra-experimental-features nix-command` support).
    # That new CLI is officially experimental (up to change),
    # but we'd be using it in way that seem extremely unlikely to break.
    # The reason is that its output will be a bit more user-friendly.
    # Perhaps use it without --print-build-logs and shown? (has color status)
    run('nix-build', srcpkg_paths[0],
        '--out-link', out_path / 'result',
        '--keep-failed',  # and keep the nix build dir for inspection
        )
    return list(out_path.glob('result*'))


def install_custom_packages(
        packages,
        **kwargs):

    cmd = ['nix-env', '--install']

    # There's no real "interactive" mode.
    dry_run = kwargs.get('interactive', False)
    if dry_run:
        cmd += ['--dry-run']

    if not len(packages) > 0:  # otherwise it tries installing everything :-)
        raise ex.InvalidInput(msg="no packages to install")
    # We don't check that `packages` are really custom packages;
    # they could be parameters, distro package names, etc.
    cmd += packages

    run(cmd, env=os.environ.copy())
